/*
* JBoss, Home of Professional Open Source.
* Copyright 2006, Red Hat Middleware LLC, and individual contributors
* as indicated by the @author tags. See the copyright.txt file in the
* distribution for a full listing of individual contributors. 
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/ 
package org.jboss.test.dependency.controller.benchmark;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import junit.framework.TestCase;

import org.jboss.dependency.util.FastCopyOnWriteSet;

/**
 * 
 * @author <a href="kabir.khan@jboss.com">Kabir Khan</a>
 * @version $Revision: 1.1 $
 */
public class TestFastCopySet extends TestCase
{
   public void testAdd() throws Throwable
   {
      FastCopyOnWriteSet<String> fastCopySet = new FastCopyOnWriteSet<String>();
      
      Map<String, Object> delegateA = fastCopySet.getDelegate(); 
      assertTrue(delegateA.isEmpty());
      
      assertTrue(fastCopySet.add("A"));
      Map<String, Object> delegateB = fastCopySet.getDelegate();
      check(delegateA);
      check(delegateB, "A");
      
      assertTrue(fastCopySet.add("B"));
      Map<String, Object> delegateC = fastCopySet.getDelegate();
      //assertTrue(delegateA.isEmpty());
      check(delegateA);
      check(delegateB, "A");
      check(delegateC, "A", "B");
      
      assertFalse(fastCopySet.add("B"));
      assertSame(delegateC, fastCopySet.getDelegate());
   }
   
   public void testAddAll() throws Throwable
   {
      FastCopyOnWriteSet<String> fastCopySet = new FastCopyOnWriteSet<String>();

      Map<String, Object> delegateA = fastCopySet.getDelegate(); 

      assertTrue(fastCopySet.addAll(createSet("A", "B")));
      Map<String, Object> delegateB = fastCopySet.getDelegate();
      check(delegateA);
      check(delegateB, "A", "B");
      
      assertTrue(fastCopySet.addAll(createSet("C", "D")));
      Map<String, Object> delegateC = fastCopySet.getDelegate();
      check(delegateA);
      check(delegateB, "A", "B");
      check(delegateC, "A", "B", "C", "D");
      
      assertTrue(fastCopySet.addAll(createSet("B", "E")));
      Map<String, Object> delegateD = fastCopySet.getDelegate();
      check(delegateA);
      check(delegateB, "A", "B");
      check(delegateC, "A", "B", "C", "D");
      check(delegateD, "A", "B", "C", "D", "E");
            
      assertFalse(fastCopySet.addAll(createSet("B", "E")));
      assertSame(delegateD, fastCopySet.getDelegate());
   }

   public void testRemove() throws Throwable
   {
      FastCopyOnWriteSet<String> fastCopySet = new FastCopyOnWriteSet<String>();
      fastCopySet.add("A");
      fastCopySet.add("B");
      fastCopySet.add("C");
      
      Map<String, Object> delegateA = fastCopySet.getDelegate(); 
      check(delegateA, "A", "B", "C");
      
      assertTrue(fastCopySet.remove("C"));
      Map<String, Object> delegateB = fastCopySet.getDelegate();
      check(delegateA, "A", "B", "C");
      check(delegateB, "A", "B");

      assertTrue(fastCopySet.remove("B"));
      Map<String, Object> delegateC = fastCopySet.getDelegate();
      check(delegateA, "A", "B", "C");
      check(delegateB, "A", "B");
      check(delegateC, "A");
      
      assertFalse(fastCopySet.remove("B"));
      assertSame(delegateC, fastCopySet.getDelegate());

      assertTrue(fastCopySet.remove("A"));
      Map<String, Object> delegateD = fastCopySet.getDelegate();
      check(delegateA, "A", "B", "C");
      check(delegateB, "A", "B");
      check(delegateC, "A");
      check(delegateD);
      
      assertFalse(fastCopySet.remove("A"));
      assertSame(delegateD, fastCopySet.getDelegate());
   }
   
   public void testRemoveAll() throws Throwable
   {
      FastCopyOnWriteSet<String> fastCopySet = new FastCopyOnWriteSet<String>();
      fastCopySet.add("A");
      fastCopySet.add("B");
      
      Map<String, Object> delegateA = fastCopySet.getDelegate();
      
      assertFalse(fastCopySet.removeAll(createSet("E", "F")));
      assertSame(delegateA, fastCopySet.getDelegate());
      
      assertTrue(fastCopySet.removeAll(createSet("A", "B")));
      Map<String, Object> delegateB = fastCopySet.getDelegate();
      check(delegateA, "A", "B");
      check(delegateB);
      
      assertFalse(fastCopySet.removeAll(createSet("A", "B")));
      assertSame(delegateB, fastCopySet.getDelegate());
   }
   
   public void testRetainAll() throws Throwable
   {
      FastCopyOnWriteSet<String> fastCopySet = new FastCopyOnWriteSet<String>();
      fastCopySet.add("A");
      fastCopySet.add("B");
      
      Map<String, Object> delegateA = fastCopySet.getDelegate();
      check(delegateA, "A", "B");
      
      assertFalse(fastCopySet.retainAll(createSet("A", "B")));
      assertSame(delegateA, fastCopySet.getDelegate());
      
      assertTrue(fastCopySet.retainAll(createSet("B", "F")));
      Map<String, Object> delegateB = fastCopySet.getDelegate();
      
      check(delegateB, "B");
      
      assertTrue(fastCopySet.retainAll(createSet("F")));
      Map<String, Object> delegateC = fastCopySet.getDelegate();
      
      check(delegateC);
   }
   
   public void testSet() throws Throwable
   {
      FastCopyOnWriteSet<String> fastCopySet = new FastCopyOnWriteSet<String>();

      assertTrue(fastCopySet.add("A"));
      check(fastCopySet, "A");
      assertFalse(fastCopySet.add("A"));
      check(fastCopySet, "A");
      
      assertTrue(fastCopySet.addAll(createSet("B", "C")));
      check(fastCopySet, "A", "B", "C");
      assertFalse(fastCopySet.addAll(createSet("B", "C")));
      check(fastCopySet, "A", "B", "C");
      
      assertTrue(fastCopySet.remove("C"));
      check(fastCopySet, "A", "B");
      assertFalse(fastCopySet.remove("C"));
      check(fastCopySet, "A", "B");

      assertFalse(fastCopySet.removeAll(createSet("C")));
      check(fastCopySet, "A", "B");
      assertTrue(fastCopySet.removeAll(createSet("A", "B", "C")));
      check(fastCopySet);
      
      assertTrue(fastCopySet.addAll(createSet("A", "B", "C")));
      check(fastCopySet, "A", "B", "C");
      assertFalse(fastCopySet.retainAll(createSet("A", "B", "C")));
      assertTrue(fastCopySet.retainAll(createSet("A", "B")));
      check(fastCopySet, "A", "B");
      assertTrue(fastCopySet.retainAll(createSet("X", "Y")));
      check(fastCopySet);
   }
   
   
   private Set<String> createSet(String...values)
   {
      Set<String> set = new HashSet<String>();
      for (String value : values)
         set.add(value);
      return set;
   }
   
   
   private void check(Map<String, ?> map, String...expected)
   {
      check(map.keySet(), expected);
   }
   
   private void check(Set<String> set, String...expected)
   {
      if (expected.length == 0)
         assertTrue(set.isEmpty());
      else
         assertFalse(set.isEmpty());
      
      assertEquals(expected.length, set.size());
      for (String s : expected)
         assertTrue(set.contains(s));
      
      Set<String> control = new HashSet<String>(Arrays.asList(expected));
      assertTrue(set.containsAll(control));
      
      Set<String> iterated = new HashSet<String>();
      for (String s : set)
         iterated.add(s);
      
      assertEquals(set, iterated);
      assertFalse(set.contains("NotThere"));
   }
}
